#FILE: blast_hit_features.pl
#AUTH: Paul Stothard (paul.stothard@gmail.com)
#DATE: July 1, 2008
#VERS: 1.0

use warnings;
use strict;
use Data::Dumper;
use File::Temp;
use Getopt::Long;
use LWP::UserAgent;
use HTTP::Request::Common;
use LWP::Simple;
use Bio::SeqIO;
use Bio::Range;
use Memoize;

#initialize options
my %options = (
    infile  => undef,
    outfile => undef,
    n       => 3,
    cache   => undef,
    verbose => undef
);

#get command line options
GetOptions(
    'i=s' => \$options{infile},
    'o=s' => \$options{outfile},
    'n=i' => \$options{n},
    'c'   => \$options{cache},
    'v'   => \$options{verbose}
);

$options{n} = get_integer( $options{n} );

#check for options
if (   !( defined( $options{infile} ) )
    or !( defined( $options{outfile} ) )
    or !( defined( $options{n} ) ) )
{
    die( print_usage() . "\n" );
}

#caching will prevent repeated downloading of the same records, but will increase memory usage.
if ( defined( $options{cache} ) ) {
    memoize('get_genbank_record');
}

open( my $OUTFILE, '>', $options{outfile} )
    or die("Cannot open file '$options{outfile}': $!");

open( my $INFILE, $options{infile} )
    or die("Cannot open file '$options{infile}': $!");

#add comments to output file
print {$OUTFILE} "#BLAST results extended using blast_hit_features.pl on "
    . get_time() . ".\n";
print {$OUTFILE}
    "#The following settings were specified for blast_hit_features.pl:\n";
foreach ( keys(%options) ) {
    if ( defined( $options{$_} ) ) {
        print {$OUTFILE} "#" . $_ . "=" . $options{$_} . "\n";
    }
}

my @columns = ();
while ( my $line = <$INFILE> ) {
    chomp($line);

#write comment lines to the output file. Comments always appear before the column titles line.
    if ( ( $line =~ m/^\#/ ) && ( scalar(@columns) == 0 ) ) {
        print {$OUTFILE} "$line\n";
    }

#the column titles line will always start with 'query_id\tmatch_id\tmatch_description\t'
    elsif (( $line =~ m/^query_id\tmatch_id\tmatch_description\t/ )
        && ( scalar(@columns) == 0 ) )
    {
        @columns = split( /\t/, $line );

        #add new column titles to the column titles line
        $line = $line
            . "\toverlapping_features\tfeatures_to_the_left\tfeatures_to_the_right";
        print {$OUTFILE} "$line\n";
    }

    else {
        print {$OUTFILE} get_hit_features( \%options, \@columns, $line )
            . "\n";
    }
}

close($INFILE) or die("Cannot close file : $!");

close($OUTFILE) or die("Cannot close file : $!");

sub get_hit_features {
    my $options = shift;
    my $columns = shift;
    my $line    = shift;

    my %hsp = ();
    my @values = split( /\t/, $line );

   #results lines without the full number of columns represent absence of hits
    if ( scalar(@values) != scalar(@$columns) ) {
        return $line;
    }
    for ( my $i = 0; $i < scalar(@$columns); $i++ ) {
        $hsp{ $$columns[$i] } = $values[$i];
    }

    print
        "Determining relationship between query $hsp{query_id} and hit $hsp{match_id}\n";

#by convention, if s_start is greater than s_end, the strand is -1. Otherwise the strand is 1.
#create a Bio::Range object to represent the position of the hit.
#http://search.cpan.org/~birney/bioperl-1.4/Bio/Range.pm
    my $range = undef;
    if ( $hsp{s_start} > $hsp{s_end} ) {
        $range = new Bio::Range(
            -start  => $hsp{s_end},
            -end    => $hsp{s_start},
            -strand => -1
        );
    }
    else {
        $range = new Bio::Range(
            -start  => $hsp{s_start},
            -end    => $hsp{s_end},
            -strand => +1
        );
    }

    #retrieve the GenBank record using the gi number if found
    my $uid = undef;
    if ( $hsp{match_id} =~ m/(ref|gi)\|(\d+)/ ) {
        $uid = $2;
    }
    else {
        print
            "Unable to find GI number in '$hsp{match_id}'--no feature information will be added for this HSP.\n";
        return $line;
    }

    my $genbank_record = get_genbank_record($uid);

    #write GenBank record to temporary file.
    my $tmp      = new File::Temp();
    my $filename = $tmp->filename;
    print {$tmp} $genbank_record;
    close($tmp) or die ("Cannot close file : $!");

    #create a Bio::SeqIO object.
    #see http://doc.bioperl.org/releases/bioperl-1.0.1/Bio/SeqIO.html
    my $seqio_object = undef;
    my $seq_object   = undef;

    $seqio_object
        = Bio::SeqIO->new( -file => $filename, -format => "genbank" );

#this step is expected to fail if the downloaded file is not a GenBank record (is an empty file for example).
    eval { $seq_object = $seqio_object->next_seq; };
    if ($@) {
        print
            "A parsable GenBank record was not obtained for hit '$hsp{match_id}'--no feature information will be added for this HSP.\n";
        return $line;
    }

    #obtain the Sequence Features associated with the sequence object.
    #see http://search.cpan.org/~birney/bioperl-1.4/Bio/SeqFeature/Generic.pm
    my @features = $seq_object->get_SeqFeatures();

    #get a list of feature objects that overlap with the hit
    my @overlapping_features = get_overlapping_features( $range, \@features );

    my @upstream_features
        = get_features_to_left( $options->{n}, $range, \@features );

    my @downstream_features
        = get_features_to_right( $options->{n}, $range, \@features );

    return $line . "\t"
        . get_context_string( $range, \@overlapping_features,
        $options->{verbose} )
        . "\t"
        . get_context_string( $range, \@upstream_features,
        $options->{verbose} )
        . "\t"
        . get_context_string( $range, \@downstream_features,
        $options->{verbose} );
}

sub get_context_string {
    my $range    = shift;
    my $features = shift;
    my $verbose  = shift;

    #number of tags to write out per feature when $verbose is not defined
    my $TAG_LIMIT = 3;
    my @output    = ();

    foreach (@$features) {
        my $feature = $_;

        #skip features of type 'source'
        if ( lc( $feature->primary_tag ) eq 'source' ) {
            next;
        }

        push( @output, "FEATURE: " . $feature->primary_tag );

     #determine relationship to $range in terms of position and add to context
        my $location = $feature->location;

        if ( $range->overlaps($location) ) {
            if ( defined($verbose) ) {
                push( @output,
                          "RELATIONSHIP WITH HSP: This feature ("
                        . $location->to_FTstring
                        . " strand="
                        . $location->strand
                        . ") overlaps with the HSP ("
                        . $range->toString()
                        . ")." );
            }
            else {
                push( @output, 'RELATIONSHIP WITH HSP: overlap' );
            }
        }
        if ( $range->start > $location->end ) {
            if ( defined($verbose) ) {
                push( @output,
                          "RELATIONSHIP WITH HSP: This feature ("
                        . $location->to_FTstring
                        . " strand="
                        . $location->strand . ") is "
                        . ( $range->start - $location->end )
                        . " bases to the left of the HSP ("
                        . $range->toString()
                        . ")." );
            }
            else {
                push( @output,
                          "RELATIONSHIP WITH HSP: "
                        . ( $range->start - $location->end )
                        . ' bp to the left' );
            }
        }
        if ( $range->end < $location->start ) {
            if ( defined($verbose) ) {
                push( @output,
                          "RELATIONSHIP WITH HSP: This feature ("
                        . $location->to_FTstring
                        . " strand="
                        . $location->strand . ") is "
                        . ( $location->start - $range->end )
                        . " bases to the right of the HSP ("
                        . $range->toString()
                        . ")." );
            }
            else {
                push( @output,
                          "RELATIONSHIP WITH HSP: "
                        . ( $location->start - $range->end )
                        . ' bp to the right' );
            }
        }

        #add tags and values to context.
        my @tag_output = ();
        my $tag_limit  = undef;
        if ( !defined($verbose) ) {
            $tag_limit = $TAG_LIMIT;
        }
        my $tag_count = 0;
        for my $tag ( $feature->get_all_tags ) {
            if ( lc($tag) eq 'translation' ) {
                next;
            }
            if ( ( defined($tag_limit) ) && ( $tag_count == $tag_limit ) ) {
                last;
            }
            push( @tag_output, "TAG: $tag" );
            $tag_count++;
            for my $value ( $feature->get_tag_values($tag) ) {
                push( @tag_output, "VALUE: $value" );
            }
        }

        push( @output, join( ', ', @tag_output ) );
    }

    my $text = join( '; ', @output );
    $text =~ s/[\t\n]//g;
    return $text;
}

sub get_overlapping_features {
    my $range                = shift;
    my $features             = shift;
    my @overlapping_features = ();

#set $use_split_ranges to 1 to examine individual ranges in split locations instead of the overall start and stop.
    my $use_split_ranges = 0;
    foreach (@$features) {
        my $feature = $_;

        #following code checks to see that
        if (   ($use_split_ranges)
            && ( $feature->location->isa('Bio::Location::SplitLocationI') ) )
        {
            my $overlaps = 0;
            for my $location ( $feature->location->sub_Location ) {
                if ( $range->overlaps($location) ) {
                    $overlaps = 1;
                    last;
                }
            }
            if ($overlaps) {
                push( @overlapping_features, $feature );
            }
        }
        else {
            my $location = $feature->location;
            if ( $range->overlaps($location) ) {
                push( @overlapping_features, $feature );
            }
        }
    }
    return @overlapping_features;
}

sub get_features_to_left {

    #note that this assumes sequence is linear
    my $n                 = shift;
    my $range             = shift;
    my $features          = shift;
    my @upstream_features = ();

    #sort features by end position, in descending order.
    @$features = map { $_->[1] }
        sort { $b->[0] <=> $a->[0] }
        map { [ $_->location->end, $_ ] } @$features;

    foreach (@$features) {
        if ( scalar(@upstream_features) == $n ) {
            last;
        }
        my $feature = $_;
        if ( $feature->location->end < $range->start ) {
            push( @upstream_features, $feature );
        }

    }

    return @upstream_features;
}

sub get_features_to_right {

    #note that this assumes sequence is linear
    my $n                   = shift;
    my $range               = shift;
    my $features            = shift;
    my @downstream_features = ();

    #sort features by start position, in ascending order.
    @$features = map { $_->[1] }
        sort { $a->[0] <=> $b->[0] }
        map { [ $_->location->start, $_ ] } @$features;

    foreach (@$features) {
        if ( scalar(@downstream_features) == $n ) {
            last;
        }
        my $feature = $_;
        if ( $feature->location->start > $range->end ) {
            push( @downstream_features, $feature );
        }

    }

    return @downstream_features;
}

sub get_genbank_record {
    my $query  = shift;
    my $url    = "http://www.ncbi.nlm.nih.gov/entrez/eutils";
    my $db     = 'nucleotide';
    my $report = 'gb';

    my $esearch = "$url/esearch.fcgi?" . "db=$db&retmax=1&usehistory=y&term=";

    my $esearch_result = get( $esearch . $query );

    $esearch_result
        =~ m/<Count>(\d+)<\/Count>.*<QueryKey>(\d+)<\/QueryKey>.*<WebEnv>(\S+)<\/WebEnv>/s;

    my $count     = $1;
    my $query_key = $2;
    my $web_env   = $3;

    my $efetch
        = "$url/efetch.fcgi?"
        . "rettype=$report&retmode=text&retstart=0&retmax=1&"
        . "db=$db&query_key=$query_key&WebEnv=$web_env";

    my $efetch_result = get($efetch);

    sleep(3);
    return $efetch_result;
}

sub get_integer {
    my $value = shift;
    my $int   = undef;
    if ( ( defined($value) ) && ( $value =~ m/(\-*\d+)/ ) ) {
        $int = $1;
    }
    return $int;
}

sub get_time {
    my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst )
        = localtime(time);
    $year += 1900;

    my @days = (
        'Sunday',   'Monday', 'Tuesday', 'Wednesday',
        'Thursday', 'Friday', 'Saturday'
    );
    my @months = (
        'January',   'February', 'March',    'April',
        'May',       'June',     'July',     'August',
        'September', 'October',  'November', 'December'
    );
    my $time
        = $days[$wday] . " "
        . $months[$mon] . " "
        . sprintf( "%02d", $mday ) . " "
        . sprintf( "%02d", $hour ) . ":"
        . sprintf( "%02d", $min ) . ":"
        . sprintf( "%02d", $sec ) . " "
        . sprintf( "%04d", $year );
    return $time;
}

sub print_usage {
    print <<BLOCK;
USAGE: perl blast_hit_features.pl [-arguments]
 -i [FILE]     : the BLAST results file to process (Required)
 -o [FILE]     : output file (Required)
 -n [INTEGER]  : the number of upstream and downstream features to
                 report. (Optional. Default is 3).
 -c            : cache downloaded GenBank files in
                 memory. (Optional. Off by default).
 -v            : provide verbose feature descriptions in
                 output. (Optional. Off by default).

 perl blast_hit_features.pl -i blastn_results.txt \
-o blastn_results_with_features.txt
BLOCK
}
